# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import warnings as wrng
from enum import Enum as enum_t
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as nmpy
import scipy.ndimage as spim
import scipy.ndimage.morphology as scph
import scipy.optimize as spop
import skimage.morphology as mrph
import skimage.segmentation as sgmt
from scipy.spatial import distance as dstc

import cell_tracking_BC.task.registration as rgst
from cell_tracking_BC.task.jaccard import PseudoJaccard


array_t = nmpy.ndarray


class compartment_t(enum_t):
    CELL = 0
    CYTOPLASM = 1
    NUCLEUS = 2


_CELL_IDX = compartment_t.CELL.value
_CYTO_IDX = compartment_t.CYTOPLASM.value


version_id_h = Tuple[int, str]
versioned_compartment_h = Union[array_t, None, version_id_h]
version_h = Tuple[versioned_compartment_h, versioned_compartment_h]
versions_h = Dict[version_id_h, version_h]

CellIssues_h = Callable[[int, array_t, dict], Optional[Union[str, Sequence[str]]]]


@dtcl.dataclass(repr=False, eq=False)
class segmentation_t:

    cell: array_t = None  # dtype=bool; Never None after instantiation
    cytoplasm: array_t = None  # dtype=bool
    # nucleus = cell - cytoplasm
    version_idx: int = 0
    version_name: str = "original"
    versions: versions_h = dtcl.field(init=False, default_factory=dict)
    invalid_cells: List[Tuple[int, Union[str, Sequence[str]]]] = dtcl.field(
        init=False, default_factory=list
    )

    # TODO: check that cytoplasm as at least one hole (and at most 2 holes?)

    def __post_init__(self) -> None:
        """"""
        version = self._NewVersionFromCompartments(self.cell, self.cytoplasm)
        self.versions[self.version] = version

    @classmethod
    def NewFromCompartments(
        cls,
        /,
        *,
        cell: array_t = None,
        cytoplasm: array_t = None,
        nucleus: array_t = None,
    ) -> segmentation_t:
        """
        Valid options:
            - cell               => cytoplasm = None, nucleus = None
            - cell, cytoplasm    => nucleus = cell - cytoplasm
            - cell, nucleus      => cytoplasm = cell - nucleus
            - cytoplasm          => cell = filled cytoplasm, nucleus = cell - cytoplasm
            - cytoplasm, nucleus => cell = cytoplasm + nucleus
        """
        if cell is not None:
            cell = cell > 0
        if cytoplasm is not None:
            cytoplasm = cytoplasm > 0
        if nucleus is not None:
            nucleus = nucleus > 0
        filled_cytoplasm = None

        if (cell is None) and (cytoplasm is None):
            raise ValueError("Cytoplasm and cell arrays both None")
        if not ((cell is None) or (cytoplasm is None) or (nucleus is None)):
            raise ValueError("Nucleus, cytoplasm and cell arrays all not None")
        if not ((cell is None) or (cytoplasm is None)):
            filled_cytoplasm = scph.binary_fill_holes(cytoplasm)
            if not nmpy.array_equal(filled_cytoplasm, cell):
                raise ValueError("Cytoplasm outer borders do not coincide with cells")
        if not ((cell is None) or (nucleus is None)):
            if nmpy.any(cell[nucleus] == False):
                raise ValueError("Nuclei outer borders not restricted to cells")
        if not ((cytoplasm is None) or (nucleus is None)):
            if nmpy.any(cytoplasm[nucleus]):
                raise ValueError("Cytoplasm and nucleus arrays intersect")
            # Necessarily not already computed above since all 3 segmentations cannot be passed
            filled_cytoplasm = scph.binary_fill_holes(cytoplasm)
            union = nmpy.logical_or(cytoplasm, nucleus)
            if not nmpy.array_equal(filled_cytoplasm, union):
                raise ValueError("Cytoplasm inner borders do not coincide with nuclei")

        if cell is None:  # Then cytoplasm is not None
            if filled_cytoplasm is None:
                if nucleus is None:
                    cell = scph.binary_fill_holes(cytoplasm)
                else:
                    cell = cytoplasm.copy()
                    cell[nucleus] = True
            else:
                cell = filled_cytoplasm
        # From then on, cell is not None
        if (cytoplasm is None) and (nucleus is not None):
            cytoplasm = cell.copy()
            cytoplasm[nucleus] = False

        return cls(cell=cell, cytoplasm=cytoplasm)

    @classmethod
    def NewFromDict(cls, dictionary: Dict[str, Any], /) -> segmentation_t:
        """"""
        instance = cls()

        instance.cell = nmpy.array(dictionary["cell"])
        instance.cytoplasm = (
            nmpy.array(dictionary["cytoplasm"])
            if dictionary["cytoplasm"] is not None
            else None
        )
        instance.version_idx = dictionary["version_idx"]
        instance.version_name = dictionary["version_name"]
        instance.versions = {}

        versions = instance.versions
        for version, values in dictionary["versions"].items():
            new_values = []
            for v_idx, value in enumerate(values):
                if (value is None) or IsVersion(value):
                    new_value = value
                else:
                    new_value = nmpy.array(value)
                new_values.append(new_value)
            versions[version] = tuple(new_values)

        return instance

    def AddFakeVersion(self, version: str, /) -> None:
        """"""
        self._AddVersionForCompartments(
            self.cell, self.cytoplasm, version, (False, False)
        )

    def _AddVersionForCompartments(
        self,
        cell: Union[array_t],
        cytoplasm: Union[array_t, None],
        version: str,
        has_changed: Tuple[bool, bool],
        /,
    ) -> None:
        """"""
        if has_changed[0]:
            self.cell = cell
        else:
            cell = self.version
        if has_changed[1]:
            self.cytoplasm = cytoplasm
        else:
            cytoplasm = self.version
        new_version = self._NewVersionFromCompartments(cell, cytoplasm)

        self.version_idx += 1
        self.version_name = version
        self.versions[self.version] = new_version

    def _NewVersionFromCompartments(
        self,
        cell: Union[array_t, version_id_h],
        cytoplasm: Union[array_t, None, version_id_h],
        /,
    ) -> version_h:
        """"""
        output = [None, None]

        if IsVersion(cell):
            cell = self._ActualOriginal(cell, _CELL_IDX)
        if IsVersion(cytoplasm):
            cytoplasm = self._ActualOriginal(cytoplasm, _CYTO_IDX)

        output[_CELL_IDX] = cell
        output[_CYTO_IDX] = cytoplasm

        return output[0], output[1]

    def _ActualOriginal(self, version: version_id_h, index: int, /) -> version_id_h:
        """"""
        output = version

        while IsVersion(self.versions[output][index]):
            output = self.versions[output][index]

        return output

    @property
    def nucleus(self) -> array_t:
        """"""
        if self.cytoplasm is None:
            raise RuntimeError(
                "Requesting nucleus map in a segmentation w/o cytoplasm map"
            )

        return nmpy.logical_xor(self.cell, self.cytoplasm)

    def NonNoneAsList(self) -> Tuple[List[array_t], Sequence[str]]:
        """
        For preparing calls to cell_tracking_BC.type.cell_t.NewFromMaps
        """
        if self.cytoplasm is None:
            segmentations = [self.cell]
            parameters = ("cell_map",)
        else:
            segmentations = [self.cytoplasm, self.cell]
            parameters = ("cytoplasm_map", "cell_map")

        return segmentations, parameters

    @property
    def version(self) -> version_id_h:
        """"""
        return self.version_idx, self.version_name

    @property
    def available_versions(
        self,
    ) -> Tuple[Tuple[compartment_t, ...], Sequence[version_id_h]]:
        """
        The sequences contain tuples (version index, version basename, version name), where version name is a
        combination of version index and basename.
        """
        if self.cytoplasm is None:
            compartments = (compartment_t.CELL,)
        else:
            compartments = (
                compartment_t.CELL,
                compartment_t.CYTOPLASM,
                compartment_t.NUCLEUS,
            )

        versions = []
        for version in self.versions.keys():
            versions.append(version)

        return compartments, versions

    def LatestCompartment(self, compartment: compartment_t, /) -> array_t:
        """"""
        if compartment is compartment_t.CELL:
            return self.cell

        if compartment is compartment_t.CYTOPLASM:
            if self.cytoplasm is not None:
                return self.cytoplasm
        elif compartment is compartment_t.NUCLEUS:
            return self.nucleus

        raise ValueError(f"{compartment}: Invalid compartment, or compartment is None")

    def CompartmentWithVersion(
        self, compartment: compartment_t, /, *, index: int = None, name: str = None
    ) -> array_t:
        """
        index: if None and name is None also, then returns latest version
        """
        if (
            ((index is None) and (name is None))
            or (index == self.version_idx)
            or (name == self.version_name)
        ):
            return self.LatestCompartment(compartment)

        if index is None:
            VersionIsAMatch = lambda _ver: name == _ver[1]
        elif name is None:
            VersionIsAMatch = lambda _ver: index == _ver[0]
        else:
            VersionIsAMatch = lambda _ver: (index, name) == _ver
        for version in self.versions.keys():
            if VersionIsAMatch(version):
                output = self._ResolvedCompartmentWithVersion(compartment, version)
                if output is None:
                    break
                else:
                    return output

        raise ValueError(
            f"{compartment}/{index}/{name}: Invalid compartment/index/name combination"
        )

    def _ResolvedCompartmentWithVersion(
        self, compartment: compartment_t, version: version_id_h, /
    ) -> Optional[array_t]:
        """"""
        if compartment is compartment_t.NUCLEUS:
            if self.cytoplasm is None:
                output = None
            else:
                cell = self.versions[version][_CELL_IDX]
                cytoplasm = self.versions[version][_CYTO_IDX]
                if IsVersion(cell):
                    cell = self.versions[cell][_CELL_IDX]
                if IsVersion(cytoplasm):
                    cytoplasm = self.versions[cytoplasm][_CYTO_IDX]
                output = nmpy.logical_xor(cell, cytoplasm)
        else:
            output = self.versions[version][compartment.value]
            if IsVersion(output):
                output = self.versions[output][compartment.value]

        return output

    @property
    def cell_areas(self) -> Sequence[int]:
        """"""
        labeled, n_cells = mrph.label(self.cell, return_num=True, connectivity=1)
        output = tuple(
            nmpy.count_nonzero(labeled == _lbl) for _lbl in range(1, n_cells + 1)
        )

        return output

    def CellAreaHistogram(
        self,
        /,
        *,
        n_bins: int = None,
        should_return_centers: bool = False,
        should_round_centers: bool = True,
    ) -> Tuple[array_t, array_t]:
        """"""
        areas = self.cell_areas

        if n_bins is None:
            n_bins = int(round(nmpy.sqrt(areas.__len__())))
        counts, edges = nmpy.histogram(areas, bins=n_bins)
        if should_return_centers:
            centers = 0.5 * (edges[:-1] + edges[1:])
            if should_round_centers:
                centers = nmpy.around(centers).astype(nmpy.uint64)
            edges = centers

        return counts, edges

    def ClearBorderObjects(self) -> None:
        """"""
        if self.cytoplasm is None:
            originals = (self.cell,)
        else:
            originals = (self.cytoplasm, self.cell)
        new_version = [_cpt.copy() for _cpt in originals]

        # --- Clear outer segmentation border objects
        has_changed = _ClearBorderObjects(new_version, originals, -1)
        if more_than_1 := (new_version.__len__() > 1):
            # --- Clear inner segmentation border objects
            has_changed |= _ClearBorderObjects(new_version, originals, 0)

        cell = new_version[-1]
        if more_than_1:
            cytoplasm = new_version[0]
            has_changed = (has_changed, has_changed)
        else:
            cytoplasm = None
            has_changed = (has_changed, False)
        self._AddVersionForCompartments(cell, cytoplasm, "cleared borders", has_changed)

    def FilterCellsOut(
        self,
        CellIssues: CellIssues_h,
        /,
        **kwargs,
    ) -> None:
        """
        Currently, only applicable to the cell segmentation when no other compartments are present

        Parameters
        ----------
        CellIssues: Arguments are: cell label (from 1), labeled segmentation, and (optional) keyword arguments; Returned
            value can be None, an str, or a sequence of str.
        kwargs: Passed to CellIsInvalid as keyword arguments
        """
        assert self.cytoplasm is None

        compartment = self.cell.copy()

        labeled, n_cells = mrph.label(compartment, return_num=True, connectivity=1)
        invalids = []
        for label in range(1, n_cells + 1):
            issues = CellIssues(label, labeled, **kwargs)
            if issues is not None:
                invalids.append((label, issues))

        if has_changed := (invalids.__len__() > 0):
            for invalid in invalids:
                compartment[labeled == invalid[0]] = 0
                self.invalid_cells.append(invalid)

        self._AddVersionForCompartments(
            compartment, self.cytoplasm, "filtered cells", (has_changed, False)
        )

    def CorrectBasedOnTemporalCoherence(
        self,
        previous: segmentation_t,
        /,
        *,
        min_jaccard: float = 0.75,
        max_area_discrepancy: float = 0.25,
        min_cell_area: int = 0,
        time_point: int = None,
    ) -> int:
        """
        min_jaccard: Actually, Pseudo-Jaccard

        Currently, only applicable to the cell segmentation when no other compartments are present
        """
        assert self.cytoplasm is None

        current = self.cell.copy()
        previous = previous.cell
        if time_point is None:
            at_time_point = ""
        else:
            at_time_point = f" at time point {time_point}"

        labeled_current, n_cells_current = mrph.label(
            current, return_num=True, connectivity=1
        )
        labeled_previous, n_cells_previous = mrph.label(
            previous, return_num=True, connectivity=1
        )
        jaccards = _PairwiseJaccards(
            n_cells_previous, n_cells_current, labeled_previous, labeled_current
        )
        c_to_p_links = _CurrentToPreviousLinks(jaccards, min_jaccard)

        n_corrections = 0
        for label_current in range(1, n_cells_current + 1):
            labels_previous = c_to_p_links.get(label_current - 1)
            if (labels_previous is not None) and (labels_previous.__len__() > 1):
                labels_previous = nmpy.array(labels_previous) + 1
                context = f"between cells {sorted(labels_previous)} and fused cell {label_current}{at_time_point}"

                where_fused = labeled_current == label_current

                split = _SplitLabels(
                    where_fused,
                    labeled_previous,
                    labels_previous,
                    max_area_discrepancy,
                    min_cell_area,
                    context,
                )
                if split is None:
                    continue

                current[where_fused] = 0
                current[split > 0] = 1
                n_corrections += 1

        self._AddVersionForCompartments(
            current,
            self.cytoplasm,
            "corrected w/ temp corr",
            (n_corrections > 0, False),
        )

        return n_corrections

    def AsDict(self) -> Dict[str, Any]:
        """"""
        output = {
            "cell": self.cell.tolist(),
            "cytoplasm": self.cytoplasm.tolist()
            if self.cytoplasm is not None
            else None,
            "version_idx": self.version_idx,
            "version_name": self.version_name,
            "versions": {},
        }

        versions = output["versions"]
        for name, values in self.versions.items():
            new_values = []
            for value in values:
                if (value is None) or IsVersion(value):
                    new_value = value
                else:
                    new_value = value.tolist()
                new_values.append(new_value)
            versions[name] = new_values

        return output


def IsVersion(maybe: Any, /) -> bool:
    """"""
    return (
        isinstance(maybe, tuple)
        and (maybe.__len__() == 2)
        and isinstance(maybe[0], int)
        and isinstance(maybe[1], str)
    )


def _ClearBorderObjects(
    compartments: List[array_t], originals: Sequence[array_t], idx: int, /
) -> bool:
    """"""
    compartment = compartments[idx]
    sgmt.clear_border(compartment, in_place=True)

    return not nmpy.array_equal(compartment, originals[idx])


def _PairwiseJaccards(
    n_cells_previous: int,
    n_cells_current: int,
    labeled_previous: array_t,
    labeled_current: array_t,
    /,
) -> array_t:
    """"""
    labels_previous = nmpy.fromiter(range(1, n_cells_previous + 1), dtype=nmpy.uint64)
    labels_current = nmpy.fromiter(range(1, n_cells_current + 1), dtype=nmpy.uint64)
    # Note the reversed parameter order in PseudoJaccard since a fusion is a division in reversed time
    _PseudoJaccard = lambda lbl_1, lbl_2: PseudoJaccard(
        labeled_current, labeled_previous, lbl_2, lbl_1
    )

    output = dstc.cdist(
        labels_previous[:, None], labels_current[:, None], metric=_PseudoJaccard
    )

    return output


def _CurrentToPreviousLinks(
    pairwise_jaccards: array_t, min_jaccard: float, /
) -> Dict[int, Sequence[int]]:
    """"""
    output = {}

    while True:
        row_idc, col_idc = spop.linear_sum_assignment(1.0 - pairwise_jaccards)
        valid_idc = pairwise_jaccards[row_idc, col_idc] > min_jaccard
        if not nmpy.any(valid_idc):
            break

        valid_row_idc = row_idc[valid_idc]
        for key, value in zip(col_idc[valid_idc], valid_row_idc):
            if key in output:
                output[key].append(value)
            else:
                output[key] = [value]

        pairwise_jaccards[valid_row_idc, :] = 0.0

    return output


def _SplitLabels(
    where_fused: array_t,
    labeled_previous: array_t,
    labels_previous: array_t,
    max_area_discrepancy: float,
    min_cell_area: int,
    context: str,
    /,
) -> Optional[array_t]:
    """"""
    output = nmpy.zeros_like(labeled_previous)
    for l_idx, previous_label in enumerate(labels_previous, start=1):
        output[labeled_previous == previous_label] = l_idx

    fused_area = nmpy.count_nonzero(where_fused)
    split_area = nmpy.count_nonzero(output)
    discrepancy = abs(split_area - fused_area) / fused_area
    if discrepancy > max_area_discrepancy:
        wrng.warn(
            f"Segmentation correction discarded due to high t-total-area/(t+1)-fused-area discrepancy "
            f"{context}: "
            f"Actual={discrepancy}; Expected<={max_area_discrepancy}"
        )
        return None

    (fused_local, split_local), corners = rgst.InCommonNonZeroRectangles(
        where_fused, output, for_rotation=True
    )

    angle = rgst.RotationInBinary(fused_local, split_local > 0)
    rotated = rgst.RotatedLabeled(split_local, angle)

    split_local = _DilatedWithLabelPreservation(rotated, fused_local, min_cell_area)
    if split_local is None:
        wrng.warn(
            f"Segmentation correction discarded due to invalid split regions "
            f"{context}"
        )
        return None

    corner = corners[0]
    rows_gbl, cols_gbl = nmpy.meshgrid(
        range(corner[0], corner[0] + split_local.shape[0]),
        range(corner[1], corner[1] + split_local.shape[1]),
        indexing="ij",
    )
    rows_lcl, cols_lcl = nmpy.indices(split_local.shape)
    valid = nmpy.logical_and(
        nmpy.logical_and(rows_gbl >= 0, cols_gbl >= 0),
        nmpy.logical_and(rows_gbl < output.shape[0], cols_gbl < output.shape[1]),
    )
    output.fill(0)
    output[rows_gbl[valid], cols_gbl[valid]] = split_local[
        rows_lcl[valid], cols_lcl[valid]
    ]

    return output


def _DilatedWithLabelPreservation(
    labeled: array_t, roi: array_t, min_cell_area: int, /
) -> Optional[array_t]:
    """"""
    output = nmpy.ones_like(labeled)

    n_labels = nmpy.amax(labeled).item()
    non_roi = nmpy.logical_not(roi)

    distance_map = spim.distance_transform_edt(labeled != 1)
    for label in range(2, n_labels + 1):
        current_map = spim.distance_transform_edt(labeled != label)
        closer_bmap = current_map < distance_map
        output[closer_bmap] = label
        distance_map[closer_bmap] = current_map[closer_bmap]

    while True:
        intermediate = nmpy.zeros_like(output)
        for label in range(1, n_labels + 1):
            where = output == label
            if not nmpy.any(where):
                return None
            # TODO: check how erosion behaves on image borders
            eroded = mrph.binary_erosion(where)
            intermediate[eroded] = label
        nmpy.copyto(output, intermediate)
        output[non_roi] = 0

        eroded_labeled, n_eroded_labels = mrph.label(
            output > 0, return_num=True, connectivity=1
        )
        areas = tuple(
            nmpy.count_nonzero(eroded_labeled == _lbl)
            for _lbl in range(1, n_eroded_labels + 1)
        )
        for label, area in enumerate(areas, start=1):
            if area < min_cell_area:
                output[eroded_labeled == label] = 0
                n_eroded_labels -= 1
        if n_eroded_labels == n_labels:
            break

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Callable, Optional, Tuple, Type, Union

import matplotlib.pyplot as pypl
import numpy as nmpy

from cell_tracking_BC.in_out.graphics.dbe.matplotlib.d_2 import (
    figure_t as figure_2d_t,
    s_drawer_2d_t,
)
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.context import DRAWING_CONTEXT
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t
from cell_tracking_BC.in_out.graphics.type.context import context_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.s_drawer_2d import (
    s_drawer_2d_t as s_drawer_t,
)
from cell_tracking_BC.type.sequence import sequence_h, sequence_t


array_t = nmpy.ndarray


def AsAnnotatedVolume(
    sequence: sequence_h,
    channel: str,
    with_segmentation: bool,
    with_cell_labels: bool,
    with_track_labels: bool,
    /,
    *,
    drawer: Union[s_drawer_t, s_drawer_2d_t] = None,
    SaveFrame: Callable[[array_t, int], None] = None,
) -> Optional[array_t]:
    """"""
    output = None  # Cannot be initialized since content (not frame) shape is unknown

    if drawer is None:
        figure, axes, drawer = FigureAxesAndDrawer(
            figure_2d_t,
            s_drawer_2d_t,
            sequence,
            channel,
            with_segmentation,
            with_cell_labels,
            with_track_labels,
            DRAWING_CONTEXT,
            True,
        )
    else:
        figure, axes = figure_2d_t.NewFigureAndAxes(offline_version=True)
        # Fix drawer
        drawer.axes = axes
        drawer.dbe = DRAWING_CONTEXT
    canvas = figure.canvas

    sequence_length = sequence.__len__()
    for time_point in range(sequence_length):
        drawer.ShowFrame(time_point=time_point, should_draw=False)
        canvas.draw()
        content = nmpy.array(canvas.renderer.buffer_rgba())[:, :, :3]
        if SaveFrame is None:
            if output is None:
                output = nmpy.empty((*content.shape, sequence_length), dtype=nmpy.uint8)
            output[..., time_point] = content
        else:
            SaveFrame(content, time_point)

    pypl.close(fig=figure)  # To prevent remaining caught in event loop

    return output


def FigureAxesAndDrawer(
    figure_class: Union[Type[figure_t], Type[figure_2d_t]],
    drawer_class: Union[Type[s_drawer_t], Type[s_drawer_2d_t]],
    sequence: sequence_t,
    channel: str,
    with_segmentation: bool,
    with_cell_labels: bool,
    with_track_labels: bool,
    dbe: context_t,
    offline: bool,
) -> Tuple[Union[figure_t, figure_2d_t], axes_2d_t, Union[s_drawer_t, s_drawer_2d_t]]:
    """"""
    figure, axes = figure_class.NewFigureAndAxes(offline_version=offline)
    drawer = drawer_class.NewForChannels(
        sequence,
        dbe,
        channel=channel,
        with_segmentation=with_segmentation,
        with_cell_labels=with_cell_labels,
        with_track_labels=with_track_labels,
        with_ticks=False,
        with_colorbar=False,
        in_axes=axes,
    )

    return figure, axes, drawer



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import matplotlib.pyplot as pypl
import numpy as nmpy
from skimage.draw import ellipse as ellipse_t

from cell_tracking_BC.type.segmentations import compartment_t, segmentations_t


MIN_JACCARD = 0.75

shape = (100, 100)

cell_1 = (10, 12, nmpy.pi / 8.0)
cell_2 = (14, 16, nmpy.pi / 10.0)
cell_3 = (15, 18, -nmpy.pi / 4.0)

position_1 = (30, 50)
shift_2 = (15, 18)
shift_3 = (23, -12)

trajectory_1_shifts = (-4, 0)
trajectory_2_shifts = (1, 4)
trajectory_3_shifts = (5, 0)
# cells_shifts = ((0,0,0),(0,0,1),(0,1,0),(1,0,0),(1,1,1))  # For checking base configurations
cells_shifts = (
    (1, 1, 1),
    (1, 0, 0),
    (1, 1, 1),
    (1, 0, 0),
    (0, 0, 0),
    (1, 1, 1),
    (1, 0, 0),
    (0, 0, 0),
    (0, 0, 1),
    (1, 1, 1),
    (0, 0, 0),
    (1, 1, 1),
)


cells = (cell_1, cell_2, cell_3)
relative_shifts = ((0, 0), shift_2, shift_3)
trajectory_shifts = (trajectory_1_shifts, trajectory_2_shifts, trajectory_3_shifts)

trajectories = tuple([] for _ in range(3))
for cells_shift in cells_shifts:
    for cell_shift, relative_shift, trajectory_shift, trajectory in zip(
        cells_shift, relative_shifts, trajectory_shifts, trajectories
    ):
        position = list(position_1)
        for c_idx in range(2):
            position[c_idx] += (
                relative_shift[c_idx] + cell_shift * trajectory_shift[c_idx]
            )
        trajectory.append(position)

frames = []
for f_idx in range(trajectories[0].__len__()):
    frame = nmpy.zeros(shape, dtype=nmpy.uint8)
    for cell, trajectory in zip(cells, trajectories):
        position = trajectory[f_idx]
        cell_map = ellipse_t(*position, *(cell[:2]), shape=shape, rotation=cell[2])
        frame[cell_map] = 1
    frames.append(frame)
uncorrected = nmpy.concatenate(frames, axis=1)

segmentations = segmentations_t.NewFromCompartmentSequences(
    frames.__len__(), cells_sgms=frames
)
segmentations.CorrectBasedOnTemporalCoherence(min_jaccard=MIN_JACCARD)
frames = tuple(_sgm.LatestCompartment(compartment_t.CELL) for _sgm in segmentations)
corrected = nmpy.concatenate(frames, axis=1)

# pypl.imsave("../../_dev/before-correction.png", uncorrected, cmap="gray")
# pypl.imsave("../../_dev/after-correction.png", corrected, cmap="gray")

pypl.matshow(uncorrected, cmap="gray")
pypl.matshow(corrected, cmap="gray")
pypl.show()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import warnings as wrng
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as nmpy
import skimage.morphology as mrph

from cell_tracking_BC.standard.uid import ShortID
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.segmentation import segmentation_t


array_t = nmpy.ndarray
inner_outer_associations_h = Dict[int, Union[int, Tuple[int, int]]]


class frame_t(array_t):
    """
    A Frame is one plane, or channel, of a (potentially multi-channel) image
    """

    def __new__(cls, array: array_t, /) -> frame_t:
        """"""
        if (n_dims := array.ndim) != 2:
            raise ValueError(
                f"{n_dims}: Invalid number of dimensions of frame with shape {array.shape}; "
                f"Expected=2=ROWSxCOLUMNS"
            )

        instance = nmpy.asarray(array).view(cls)
        instance.path = None
        instance.cells = []
        instance.runtime = {}

        return instance

    def __array_finalize__(self, array: Optional[array_t], /) -> None:
        """"""
        if array is None:
            return
        self.path = getattr(array, "path", None)
        self.cells = getattr(array, "cells", None)

    def ApplyTransform(self, Transform: transform_h, /, **kwargs) -> None:
        """"""
        transformed = Transform(self, **kwargs)
        self[...] = transformed[...]

    def AddCellsFromSegmentation(
        self,
        segmentation: segmentation_t,
    ) -> None:
        """
        Segmentation are supposed to be binary (as opposed to already labeled)
        """
        # --- Generic inner and outer segmentation list for all options
        non_nones, parameters = segmentation.NonNoneAsList()

        # --- Segmentation labeling and (if needed) inner-outer label associations
        labeleds = []
        n_objects_per_sgm = []
        for non_none in non_nones:
            labeled, n_objects = mrph.label(non_none, return_num=True, connectivity=1)
            labeleds.append(labeled)
            n_objects_per_sgm.append(n_objects)
        if labeleds.__len__() > 1:
            inner_label_of_outer = _InnerOuterAssociations(labeleds, n_objects_per_sgm)
            _CheckInnerOuterAssociations(inner_label_of_outer, n_objects_per_sgm)
        else:
            inner_label_of_outer = None

        # --- Cell creation and addition
        for outer_label in range(1, n_objects_per_sgm[-1] + 1):
            if inner_label_of_outer is None:
                labels = (outer_label,)
            else:
                labels = (inner_label_of_outer[outer_label], outer_label)
            kwargs = {}
            additional_nucleus = None
            for parameter, labeled, label in zip(parameters, labeleds, labels):
                label: Union[int, Tuple[int, int]]
                if isinstance(label, int):
                    kwargs[parameter] = labeled == label
                else:
                    kwargs[parameter] = labeled == label[0]
                    additional_nucleus = label[1]
            cell = cell_t.NewFromMaps(labels[-1], **kwargs)
            if additional_nucleus is not None:
                cell.AddNucleus(labeleds[0] == additional_nucleus)
            self.cells.append(cell)

    def __str__(self) -> str:
        """"""
        return (
            f"{self.__class__.__name__.upper()}.{ShortID(id(self))}:\n"
            f"    {self.path=}\n"
            f"    {self.shape=}\n"
            f"    {self.cells.__len__()=}"
        )


transform_h = Callable[[frame_t, Dict[str, Any]], array_t]


def _InnerOuterAssociations(
    segmentations: List[array_t], n_objects_per_sgm: Sequence[int], /
) -> inner_outer_associations_h:
    """"""
    output = {}

    for inner_label in range(1, n_objects_per_sgm[0] + 1):
        dilated = mrph.dilation(segmentations[0] == inner_label)
        outer_label = nmpy.amax(segmentations[1][dilated])
        if outer_label in output:
            output[outer_label] = (
                output[outer_label],
                inner_label,
            )
            if output[outer_label].__len__() > 2:
                raise RuntimeError("Cell with more than 2 nuclei")
        else:
            output[outer_label] = inner_label

    return output


def _CheckInnerOuterAssociations(
    inner_label_of_outer: inner_outer_associations_h,
    n_objects_per_sgm: Sequence[int],
    /,
) -> None:
    """"""
    if n_objects_per_sgm[0] != n_objects_per_sgm[1]:
        wrng.warn(
            f"Mismatch in number of segmented objects: {n_objects_per_sgm}; "
            f"Might be due to cell divisions"
        )

    inner_labels = []
    for label in inner_label_of_outer.values():
        if isinstance(label, int):
            inner_labels.append(label)
        else:
            inner_labels.extend(label)
    outer_labels = tuple(inner_label_of_outer.keys())
    n_inner_labels = nmpy.unique(inner_labels).size
    n_outer_labels = nmpy.unique(outer_labels).size
    if (n_missing_associations := n_objects_per_sgm[0] - n_inner_labels) > 0:
        wrng.warn(
            f"{n_missing_associations} inner comportment(s) not associated with outer compartments"
        )
    if (n_missing_associations := n_objects_per_sgm[1] - n_outer_labels) > 0:
        wrng.warn(
            f"{n_missing_associations} outer comportment(s) without inner compartment(s)"
        )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Sequence, Tuple, Union

import matplotlib.projections as prjs
import matplotlib.pyplot as pypl
from matplotlib import rc as SetMatplotlibConfig
from matplotlib import transforms as trsf
from matplotlib.backend_bases import MouseEvent as mouse_event_t
from matplotlib.backends.backend_agg import FigureCanvasAgg as agg_canvas_t
from matplotlib.image import AxesImage as axes_image_t
from matplotlib.text import Annotation as annotation_t
from numpy import ndarray as array_t

from cell_tracking_BC.in_out.graphics.dbe.matplotlib.d_any import axes_t as axes_anyd_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.d_any import (
    figure_t as figure_anyd_t,
)
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.style import (
    CELL_ANNOTATION_BBOX_STYLE_DEFAULT,
    CELL_ANNOTATION_BBOX_STYLE_HIGHLIGHT,
    CELL_ANNOTATION_COLOR_DEFAULT,
    CELL_ANNOTATION_COLOR_HIGHLIGHT,
)
from cell_tracking_BC.in_out.graphics.type.annotation import cell_annotation_h
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t as base_axes_2d_t
from cell_tracking_BC.in_out.graphics.type.s_drawer_2d import (
    s_drawer_2d_t as base_s_drawer_2d_t,
)
from cell_tracking_BC.in_out.graphics.type.s_viewer_2d import (
    s_viewer_2d_t as base_s_viewer_2d_t,
)
from cell_tracking_BC.in_out.graphics.type.t_viewer_2d import (
    t_viewer_2d_t as base_t_viewer_2d_t,
)


pypl_axes_t = pypl.Axes


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_t(axes_anyd_t, base_axes_2d_t):

    name = "ctBC_two_d_axes"
    __image__: axes_image_t = None
    PlotPoints = pypl_axes_t.scatter
    PlotLines = pypl_axes_t.plot
    PlotText = pypl_axes_t.text
    PlotAnnotation = pypl_axes_t.annotate

    def SetTimeAxisProperties(self, latest: int, /) -> None:
        """"""
        self.set_xlim(0, latest)
        self.set_xticks(range(latest + 1))
        self.set_xticklabels(str(_idx) for _idx in range(latest + 1))

    def SetTrackingAxesProperties(self, tick_positions: Sequence[float], /) -> None:
        """"""
        self.set_xlabel("time points")
        self.set_ylabel("tracks")
        self.yaxis.set_label_position("right")
        self.yaxis.tick_right()
        self.yaxis.set_ticks(tick_positions)

    def PlotCellAnnotation(
        self, position: Tuple[float, float], text: cell_annotation_h, /, **kwargs
    ) -> Union[annotation_t, Sequence[annotation_t]]:
        """"""
        if isinstance(text, str):
            output = self.annotate(
                text,
                position,
                ha="center",
                **kwargs,
            )
        else:
            output = []

            renderer = self.figure.canvas.get_renderer()
            transform = self.transData
            for piece, properties in text:
                properties = {**properties, **kwargs}
                piece = self.text(*position, piece, transform=transform, **properties)
                piece.draw(renderer)
                extent = piece.get_window_extent()
                transform = trsf.offset_copy(
                    piece.get_transform(), x=extent.width, units="dots"
                )

                output.append(piece)

        return output

    def PlotImage(
        self, image: array_t, /, *, interval: Tuple[float, float] = None
    ) -> None:
        """"""
        pypl_image = self.matshow(image, cmap="gray")
        if interval is not None:
            pypl_image.set_clim(*interval)

        self.__image__ = pypl_image

    def UpdateImage(
        self,
        picture: array_t,
        /,
        *,
        interval: Tuple[float, float] = None,
        should_update_limits: bool = False,
    ) -> None:
        """"""
        self.__image__.set_array(picture)
        if should_update_limits:
            self.__image__.set_clim(*interval)


prjs.register_projection(axes_t)


@dtcl.dataclass(init=False, repr=False, eq=False)
class figure_t(figure_anyd_t):
    @classmethod
    def NewFigureAndAxes(
        cls,
        /,
        *,
        n_rows: int = 1,
        n_cols: int = 1,
        title: str = None,
        offline_version: bool = False,
    ) -> Tuple[figure_t, Union[axes_t, Sequence[axes_t], Sequence[Sequence[axes_t]]],]:
        if offline_version:
            figure = pypl.Figure()
            _ = agg_canvas_t(
                figure
            )  # This also sets the figure canvas (see Matplotlib source)
            axes = figure.subplots(
                nrows=n_rows,
                ncols=n_cols,
                subplot_kw={"projection": "ctBC_two_d_axes"},
            )
            figure.Close = lambda: pypl.close(fig=figure)
        else:
            figure, axes = pypl.subplots(
                nrows=n_rows,
                ncols=n_cols,
                FigureClass=cls,
                subplot_kw={"projection": "ctBC_two_d_axes"},
            )

        if title is not None:
            first_axes = axes
            while isinstance(first_axes, Sequence):
                first_axes = first_axes[0]
            first_axes.set_title(title)

        return figure, axes


@dtcl.dataclass(repr=False, eq=False)
class s_drawer_2d_t(base_s_drawer_2d_t):
    def HighlightAnnotation(self, label: int, /, *, should_draw: bool = True) -> None:
        """"""
        pass

    def AddColorbarForImage(self) -> None:
        """"""
        self.figure.colorbar(self.axes.__image__, ax=self.axes)


@dtcl.dataclass(repr=False, eq=False)
class s_viewer_2d_t(base_s_viewer_2d_t):
    def HighlightAnnotation(self, label: int, /, *, should_draw: bool = True) -> None:
        """"""
        if label == self.current_label:
            return

        for which, what in self.annotations:
            if not isinstance(what, Sequence):
                what = (what,)
            if label == which:
                color = CELL_ANNOTATION_COLOR_HIGHLIGHT
                bbox = CELL_ANNOTATION_BBOX_STYLE_HIGHLIGHT
            else:
                color = CELL_ANNOTATION_COLOR_DEFAULT
                bbox = CELL_ANNOTATION_BBOX_STYLE_DEFAULT
            for text in what:
                text.set_color(color)
                if "\n" in text.get_text():
                    text.set_bbox(bbox)

        self.current_label = label

        if should_draw:
            self.figure.canvas.draw_idle()

    def AddColorbarForImage(self) -> None:
        """"""
        self.figure.colorbar(self.axes.__image__, ax=self.axes)

    def _ActivateEventProcessing(self, more_than_one: bool, /) -> None:
        """"""
        SetMatplotlibConfig("keymap", save=[])
        super()._ActivateEventProcessing(more_than_one)


@dtcl.dataclass(repr=False, eq=False)
class t_viewer_2d_t(base_t_viewer_2d_t):
    def _ShowAnnotation(
        self,
        event: mouse_event_t,
        /,
    ) -> None:
        """"""
        inside, details = self.scatter.contains(event)
        if inside:
            idx = details["ind"][0]
            time_point = self.time_points[idx]
            label = self.labels[idx]

            position = self.scatter.get_offsets()[idx]
            text = f"Time {time_point}\nCell {label}\nAffty {self.affinities[idx]:.2f}"
            self.annotation.xy = position
            self.annotation.set_text(text)
            self.annotation.set_visible(True)

            self.viewer.ShowFrame(time_point=time_point, highlighted=label)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from abc import ABC as abc_t
from abc import abstractmethod
from typing import Any, Callable, ClassVar, Dict, Iterable, Iterator, List
from typing import Optional, Sequence, Tuple, Union

import networkx as grph

from cell_tracking_BC.standard.issue import ISSUE_SEPARATOR
from cell_tracking_BC.standard.number import MAX_INT
from cell_tracking_BC.standard.uid import ShortID
from cell_tracking_BC.type.cell import cell_t, state_t


TIME_POINT = "time_point"  # Leave it here (as opposed to making it a class variable) since it is used "everywhere"


class unstructured_track_t(grph.DiGraph):
    in_degree: Callable[[cell_t], int]
    out_degree: Callable[[cell_t], int]

    def Issues(self) -> Optional[Sequence[str]]:
        """"""
        output = []

        for cell in self.nodes:
            if (n_predecessors := self.in_degree(cell)) > 1:
                output.append(
                    f"Cell {cell.label}{ISSUE_SEPARATOR}{n_predecessors} predecessors. Expected=0 or 1."
                )

        if output.__len__() == 0:
            output = None

        return output

    def RootCellWithTimePoint(self) -> Tuple[cell_t, int]:
        """"""
        output = tuple(
            _rcd for _rcd in self.nodes.data(TIME_POINT) if self.in_degree(_rcd[0]) == 0
        )

        if (n_roots := output.__len__()) != 1:
            raise ValueError(f"{n_roots}: Invalid number of root cells; Expected=1")

        return output[0]

    def LeafCellsWithTimePoints(self) -> Tuple[Sequence[cell_t], Sequence[int]]:
        """"""
        # TODO: Contact the Networkx team about the following comment (or check code on github)
        #     /!\ It seems that networkx.DiGraph.nodes.data does not guarantee the node enumeration order. This could be
        #     inconvenient for reproducibility checks.
        records = (
            _rcd
            for _rcd in self.nodes.data(TIME_POINT)
            if self.out_degree(_rcd[0]) == 0
        )

        leaves, time_points = zip(*records)
        leaves = tuple(leaves)
        time_points = tuple(time_points)

        return leaves, time_points

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        time_points = grph.get_node_attributes(self, TIME_POINT)

        for edge in self.edges:
            time_point = time_points[edge[0]]
            is_last = self.out_degree(edge[1]) == 0
            yield time_point, *edge, is_last


@dtcl.dataclass(init=False, repr=False, eq=False)
class structured_track_t(abc_t):
    root: cell_t = None
    leaves: Sequence[cell_t] = None

    @property
    @abstractmethod
    def cells(self) -> Union[Iterable, Iterator]:
        """"""

    @property
    @abstractmethod
    def labels(self) -> Sequence[int]:
        """
        Single track labels
        """

    @property
    def labels_as_str(self) -> str:
        """
        Single track labels
        """
        return "+".join(str(_lbl) for _lbl in sorted(self.labels))

    @property
    def n_leaves(self) -> int:
        """"""
        return self.leaves.__len__()

    @property
    def lengths(self) -> Sequence[int]:
        """
        Segment-wise, not node-wise
        """
        return tuple(_ltp - self.root_time_point for _ltp in self.leaves_time_points)

    @property
    @abstractmethod
    def root_time_point(self) -> int:
        """"""

    @property
    @abstractmethod
    def leaves_time_points(self) -> Sequence[int]:
        """"""

    @abstractmethod
    def CellTimePoint(self, cell: cell_t) -> int:
        """"""

    @abstractmethod
    def CellSuccessors(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[cell_t]]:
        """
        Accounts for pruning
        """

    @abstractmethod
    def DividingCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""

    @property
    @abstractmethod
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""

    @abstractmethod
    def Pieces(
        self, /, *, from_cell: cell_t = None, with_time_point: int = None
    ) -> Sequence[single_track_t]:
        """"""

    @property
    def single_tracks_iterator(self) -> Iterator[single_track_t]:
        """"""
        for leaf in self.leaves:
            yield self.TrackWithLeaf(leaf, check=False)

    @abstractmethod
    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """"""

    @abstractmethod
    def TrackLabelWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[int]:
        """"""

    @abstractmethod
    def TrackWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[single_track_t]:
        """"""

    @abstractmethod
    def AsSingleTrack(self) -> single_track_t:
        """"""

    def Prune(
        self,
        ShouldStartPruning: Callable[[int, cell_t, dict], bool],
        /,
        *,
        ShouldStopPruning: Callable[[int, cell_t, dict], bool] = None,
        start_arguments: Dict[str, Any] = None,
        stop_arguments: Dict[str, Any] = None,
        called_from_educated_code: bool = False,
    ) -> None:
        """
        ShouldSt*Pruning arguments: time point, cell, kwargs

        called_from_educated_code: A track might be entirely pruned out here. Since it cannot remove itself from its
        referring objects, this method must be called from a piece of code that can do it if needed. This parameter is
        meant to prevent an "accidental" call from some "uneducated" piece of code.
        """
        if not called_from_educated_code:
            raise ValueError(
                f"{self.Prune.__name__}: Must be called from a piece of code handling full pruning"
            )

        if start_arguments is None:
            start_arguments = {}
        if (ShouldStopPruning is not None) and (stop_arguments is None):
            stop_arguments = {}

        dividing_cells = self.DividingCells()
        for single in self.single_tracks_iterator:
            t_idx = 0

            pruning = ShouldStopPruning is not None
            while pruning and (t_idx < single.__len__()):
                cell = single[t_idx]
                if (cell in dividing_cells) or ShouldStopPruning(
                    t_idx, cell, **stop_arguments
                ):
                    # Pruning stops at the first dividing cell since pruning further would then make several tracks out
                    # of one.
                    pruning = False
                    if t_idx > 0:
                        self.root = cell
                else:
                    cell.state = state_t.pruned
                    t_idx += 1

            while t_idx < single.__len__():
                if ShouldStartPruning(t_idx, single[t_idx], **start_arguments):
                    for cell in single[t_idx:]:
                        cell.state = state_t.pruned
                    break
                t_idx += 1

        if not self.IsFullyPruned():
            leaves = []
            # Adding Iterable to the class inheritance silences this warning. However, forking tracks become
            # un-instantiable for it lacks an __iter__ method.
            for cell in self:
                if cell.state.IsActive() and (self.CellSuccessors(cell).__len__() == 0):
                    leaves.append(cell)
            self.leaves = leaves

    def IsFullyPruned(self) -> bool:
        """"""
        return self.root.state == state_t.pruned

    def __str__(self) -> str:
        """"""
        if hasattr(self, "nodes"):
            cells = self.nodes
        else:
            cells = self
        cell_labels = tuple(_cll.label for _cll in cells)

        return (
            f"{self.__class__.__name__.upper()}.{ShortID(id(self))}:\n"
            f"    {self.labels=}\n"
            f"    {self.root_time_point=}\n"
            f"    {self.leaves_time_points=}\n"
            f"    {self.lengths=}\n"
            f"    {cell_labels}"
        )


@dtcl.dataclass(init=False, repr=False, eq=False)
class single_track_t(structured_track_t, List[cell_t]):
    _label: int = None  # Single track label
    _root_time_point: int = None
    affinities: Sequence[float] = None

    def __init__(self, *args, **kwargs) -> None:
        """"""
        list.__init__(self, *args, **kwargs)

    @classmethod
    def NewFromOrderedCells(
        cls,
        cells: Sequence[cell_t],
        affinities: Sequence[float],
        root_time_point: int,
        label: Optional[int],
        /,
    ) -> single_track_t:
        """
        This must be the only place where direct instantiation is allowed. Anywhere else, instantiation must be
        performed with this class method.

        label: Can be None only to accommodate the creation of branches as single tracks
        """
        instance = cls(cells)

        instance._label = label
        instance.root = instance[0]
        instance.leaves = (instance[-1],)
        instance._root_time_point = root_time_point
        instance.affinities = affinities

        return instance

    @property
    def cells(self) -> Union[Iterable, Iterator]:
        """"""
        return self

    @property
    def label(self) -> int:
        """"""
        return self._label

    @property
    def labels(self) -> Sequence[int]:
        """"""
        return (self.label,)

    @property
    def length(self) -> int:
        """
        Segment-wise, not node-wise
        """
        return self.lengths[
            0
        ]  # Do not use list length as it does not account for pruned parts

    @property
    def root_time_point(self) -> int:
        """"""
        return self._root_time_point

    @property
    def leaf(self) -> cell_t:
        """"""
        return self.leaves[0]

    @property
    def leaf_time_point(self) -> int:
        """"""
        return self.CellTimePoint(self.leaf)

    @property
    def leaves_time_points(self) -> Sequence[int]:
        """"""
        return (self.leaf_time_point,)

    def CellTimePoint(self, cell: cell_t) -> int:
        """"""
        return self.root_time_point + self.index(cell) - self.index(self.root)

    def CellSuccessors(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[cell_t]]:
        """"""
        if (not check) or (cell in self):
            where = self.index(cell)
            if where < self.__len__() - 1:
                successor = self[where + 1]
                if successor.state.IsActive():
                    output = (successor,)
                else:
                    output = ()
            else:
                output = ()

            return output

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def DividingCells(
        self, /, *, _: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        return ()

    def _unpruned_piece(self) -> single_track_t:
        """"""
        root_idx = self.index(self.root)
        leaf_idx = self.index(self.leaf)

        # Just slicing returns an object of type list
        output = self.__class__(self[root_idx : (leaf_idx + 1)])
        for field in dtcl.fields(self):
            setattr(output, field.name, getattr(self, field.name))

        return output

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        unpruned = self._unpruned_piece()

        n_cells = unpruned.__len__()
        for c_idx in range(1, n_cells):
            time_point = self.root_time_point + c_idx - 1
            is_last = c_idx == n_cells - 1
            yield time_point, *unpruned[(c_idx - 1) : (c_idx + 1)], is_last

    def Pieces(self, /, **_) -> Sequence[single_track_t]:
        """"""
        return (self._unpruned_piece(),)

    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """"""
        if (not check) or (cell in self):
            return self.labels

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def TrackLabelWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[int]:
        """"""
        if (not check) or (leaf in self.leaves):
            return self.label

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def TrackWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[single_track_t]:
        """"""
        if (not check) or (leaf in self.leaves):
            return self._unpruned_piece()

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def AsSingleTrack(self) -> single_track_t:
        """"""
        return self._unpruned_piece()

    def AsRowsColsTimes(
        self, /, *, with_labels: bool = False
    ) -> Union[
        Tuple[Tuple[float, ...], Tuple[float, ...], Tuple[int, ...]],
        Tuple[Tuple[float, ...], Tuple[float, ...], Tuple[int, ...], Tuple[int, ...]],
    ]:
        """"""
        unpruned = self._unpruned_piece()
        n_cells = unpruned.__len__()

        rows, cols = tuple(zip(*(_cll.centroid.tolist() for _cll in unpruned)))
        times = tuple(range(self.root_time_point, self.root_time_point + n_cells))

        if with_labels:
            labels = tuple(_cll.label for _cll in unpruned)
            return rows, cols, times, labels

        return rows, cols, times


# Cannot be a dataclass due to in/out_degree declarations (which are only here to silence unfound attribute warnings)
class forking_track_t(structured_track_t, grph.DiGraph):
    """
    Affinities are stored as edge attributes
    """

    SINGLE_TRACK_LABEL: ClassVar[str] = "single_track_label"

    in_degree: Callable[[cell_t], int]
    out_degree: Callable[[cell_t], int]

    def __init__(self, *args, **kwargs) -> None:
        """"""
        grph.DiGraph.__init__(self, *args, **kwargs)

    @classmethod
    def NewFromUnstructuredTrack(
        cls, track: unstructured_track_t, next_single_track_label: int, /
    ) -> Tuple[forking_track_t, int]:
        """"""
        instance = cls(track)

        instance.root, _ = track.RootCellWithTimePoint()
        instance.leaves, _ = track.LeafCellsWithTimePoints()

        for label, leaf in enumerate(instance.leaves, start=next_single_track_label):
            # Adds attribute "forking_track_t.SINGLE_TRACK_LABEL" with value "label"
            # to leaf node indexed by "leaf" (not to the leaf itself).
            grph.set_node_attributes(
                instance,
                {leaf: label},
                name=forking_track_t.SINGLE_TRACK_LABEL,
            )

        return instance, next_single_track_label + instance.n_leaves

    @property
    def cells(self) -> Union[Iterable, Iterator]:
        """"""
        return self.nodes

    @property
    def labels(self) -> Sequence[int]:
        """"""
        output = []

        for leaf in self.leaves:
            label = self.nodes[leaf].get(forking_track_t.SINGLE_TRACK_LABEL)
            if label is None:
                current = leaf
                neighbors = tuple(self.neighbors(current))
                while neighbors.__len__() > 0:
                    current = neighbors[0]
                    neighbors = tuple(self.neighbors(current))
                label = self.nodes[current][forking_track_t.SINGLE_TRACK_LABEL]

            output.append(label)

        return output

    @property
    def root_time_point(self) -> int:
        """"""
        return self.CellTimePoint(self.root)

    @property
    def leaves_time_points(self) -> Sequence[int]:
        """"""
        return tuple(self.CellTimePoint(_lef) for _lef in self.leaves)

    @property
    def affinities(self) -> Sequence[float]:
        """"""
        output = []

        for piece in self.Pieces():
            output.extend(piece.affinities)

        return output

    def CellTimePoint(self, cell: cell_t) -> int:
        """"""
        return self.nodes[cell][TIME_POINT]

    def CellSuccessors(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[cell_t]]:
        """"""
        if (not check) or (cell in self):
            return tuple(_ngh for _ngh in self.neighbors(cell) if _ngh.state.IsActive())

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def DividingCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        if with_time_point:
            output = (
                _rcd
                for _rcd in self.nodes.data(TIME_POINT)
                if (self.out_degree(_cll := _rcd[0]) > 1) and _cll.state.IsActive()
            )
        else:
            output = (
                _cll
                for _cll in self.nodes
                if (self.out_degree(_cll) > 1) and _cll.state.IsActive()
            )

        return tuple(output)

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        time_points = grph.get_node_attributes(self, TIME_POINT)

        for edge in self.edges:
            if all(_cll.state.IsActive() for _cll in edge):
                time_point = time_points[edge[0]]
                is_last = edge[1] in self.leaves
                yield time_point, *edge, is_last

    def Pieces(
        self, /, *, from_cell: cell_t = None, with_time_point: int = None
    ) -> Sequence[single_track_t]:
        """"""
        output = []

        if from_cell is None:
            piece = [self.root]
            root_time_point = self.root_time_point
        else:
            if not from_cell.state.IsActive():
                raise ValueError(f"{from_cell}: Not an active cell")
            piece = [from_cell]
            root_time_point = with_time_point
        affinities = []

        while True:
            last_cell = piece[-1]

            if last_cell in self.leaves:
                neighbors = None
                n_neighbors = 0
            else:
                neighbors = self.CellSuccessors(last_cell, check=False)
                n_neighbors = neighbors.__len__()

            if n_neighbors == 0:
                label = self.TrackLabelWithLeaf(last_cell, check=False)
                output.append(
                    single_track_t.NewFromOrderedCells(
                        piece, affinities, root_time_point, label
                    )
                )
                break
            elif n_neighbors == 1:
                next_cell = neighbors[0]
                piece.append(next_cell)
                affinities.append(self[last_cell][next_cell]["affinity"])
            else:
                output.append(
                    single_track_t.NewFromOrderedCells(
                        piece, affinities, root_time_point, None
                    )
                )
                next_time_point = root_time_point + piece.__len__()
                for neighbor in neighbors:
                    pieces = self.Pieces(
                        from_cell=neighbor,
                        with_time_point=next_time_point,
                    )
                    for piece in pieces:
                        if piece[0] is neighbor:
                            cells = (last_cell,) + tuple(piece)
                            affinity = self[last_cell][neighbor]["affinity"]
                            affinities = (affinity,) + tuple(piece.affinities)
                            piece = single_track_t.NewFromOrderedCells(
                                cells,
                                affinities,
                                piece.root_time_point - 1,
                                piece.label,
                            )
                        output.append(piece)
                break

        return output

    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """"""
        output = []

        if check and not cell.state.IsActive():
            raise ValueError(f"{cell}: Not an active cell")

        for leaf in self.leaves:
            try:
                _ = grph.shortest_path(self, source=cell, target=leaf)
            except grph.NetworkXNoPath:
                continue
            output.append(self.nodes[leaf][forking_track_t.SINGLE_TRACK_LABEL])

        if (not check) or (output.__len__() > 0):
            return output

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def TrackLabelWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[int]:
        """
        TrackLabelWithLeaf: Implicitly, it is SingleTrackLabelWithLeaf
        """
        if (not check) or (leaf in self.leaves):
            return self.nodes[leaf][forking_track_t.SINGLE_TRACK_LABEL]

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def TrackWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[single_track_t]:
        """
        TrackWithLeaf: Implicitly, it is SingleTrackWithLeaf
        """
        if (not check) or (leaf in self.leaves):
            cells = grph.shortest_path(self, source=self.root, target=leaf)
            affinities = tuple(
                self[cells[_idx]][cells[_idx + 1]]["affinity"]
                for _idx in range(cells.__len__() - 1)
            )
            label = self.TrackLabelWithLeaf(leaf, check=False)
            output = single_track_t.NewFromOrderedCells(
                cells, affinities, self.root_time_point, label
            )

            return output

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def AsSingleTrack(self) -> single_track_t:
        """"""
        output = [self.root]

        affinities = []
        while True:
            last_cell = output[-1]

            if last_cell in self.leaves:
                neighbors = None
                n_neighbors = 0
            else:
                neighbors = self.CellSuccessors(last_cell, check=False)
                n_neighbors = neighbors.__len__()

            if n_neighbors == 0:
                label = self.TrackLabelWithLeaf(last_cell, check=False)
                break
            elif n_neighbors == 1:
                next_cell = neighbors[0]
                output.append(next_cell)
                affinity = self[last_cell][next_cell]["affinity"]
                affinities.append(affinity)
            else:
                raise ValueError(
                    f"Attempt to convert the forking track with root {self.root} and "
                    f"{self.n_leaves} leaves into a single track"
                )

        output = single_track_t.NewFromOrderedCells(
            output, affinities, self.root_time_point, label
        )

        return output


structured_track_h = Union[single_track_t, forking_track_t]
any_track_h = Union[unstructured_track_t, structured_track_h]


def DivisionTimePoints(
    dividing_cells: Sequence[Tuple[cell_t, int]], /
) -> Tuple[Optional[Sequence[int]], int]:
    """"""
    division_time_points = tuple(_elm[1] for _elm in dividing_cells)
    if division_time_points.__len__() > 0:
        last_div_frm = division_time_points[-1] + 1
    else:
        division_time_points = None  # Used to be (-1,)
        last_div_frm = 0

    return division_time_points, last_div_frm


def BasicTrackIssues(
    track: structured_track_h,
    /,
    *,
    root_time_point_interval: Sequence[Optional[int]] = (0, 0),
    leaves_time_point_intervals: Sequence[Optional[Sequence[Optional[int]]]] = (
        (1, None),
        (1, None),
    ),
    min_lengths: Sequence[Optional[int]] = (1, 1),
    max_n_children: int = 2,
    can_touch_border: bool = False,
) -> Optional[Sequence[str]]:
    """
    All parameters: any limit can be ignored by setting it to None
    All intervals are inclusive.
    leaf_time_point_intervals and min_lengths: first element is for the shortest branch, the second is for the longest.
        For single tracks, both are the same.
    min_lengths: edge-wise lengths, inclusive
    max_n_children: inclusive
    """
    output = []

    if track.root.state == state_t.dead:
        output.append('Root cell has a "dead" state')

    mini, maxi = _IntervalWithDefaults(root_time_point_interval, 0, MAX_INT)
    if not (mini <= track.root_time_point <= maxi):
        output.append(
            f"{track.root_time_point}{ISSUE_SEPARATOR}Invalid root time point. Expected={mini}..{maxi}"
        )

    min_ltp = min(track.leaves_time_points)
    max_ltp = max(track.leaves_time_points)
    for value, ltp_interval, which in zip(
        (min_ltp, max_ltp), leaves_time_point_intervals, ("shortest", "longest")
    ):
        if ltp_interval is None:
            continue
        mini, maxi = _IntervalWithDefaults(ltp_interval, 0, MAX_INT)
        if not (mini <= value <= maxi):
            output.append(
                f"{value}{ISSUE_SEPARATOR}Invalid leaf time point of {which} branch. Expected={mini}..{maxi}"
            )

    min_lgh = track.lengths[0]
    max_lgh = track.lengths[-1]
    # _IntervalWithDefaults is used on a non-interval!
    min_lengths = _IntervalWithDefaults(min_lengths, 0, 0)
    for value, mini, which in zip(
        (min_lgh, max_lgh), min_lengths, ("shortest", "longest")
    ):
        if value < mini:
            output.append(
                f"{value}{ISSUE_SEPARATOR}Invalid (edge-wise) length of {which} branch. Expected>={mini}"
            )

    if isinstance(track, forking_track_t):
        for cell in track.nodes:
            if (n_children := track.out_degree(cell)) > max_n_children:
                output.append(
                    f"C{cell.label}T{track.CellTimePoint(cell)}{ISSUE_SEPARATOR}"
                    f"{n_children} successors. Expected=0..{max_n_children}"
                )

    if not can_touch_border:
        for cell in track.cells:
            if cell.touches_border:
                output.append(
                    f"C{cell.label}T{track.CellTimePoint(cell)}{ISSUE_SEPARATOR}Touches frame border"
                )

    if output.__len__() == 0:
        output = None

    return output


def _IntervalWithDefaults(
    interval: Sequence[Optional[int]], default_min: int, default_max: int, /
) -> Sequence[int]:
    """"""
    tvl_min, tvl_max = interval
    if tvl_min is None:
        tvl_min = default_min
    if tvl_max is None:
        tvl_max = default_max

    return tvl_min, tvl_max



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import warnings as wrng
from sys import maxsize as MAX_INTEGER
from typing import List, Optional, Sequence, Tuple, cast

import networkx as grph
import numpy as nmpy

from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.type.annotation import annotation_h
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t as axes_t
from cell_tracking_BC.in_out.graphics.type.color import colormap_h, rgba_color_h
from cell_tracking_BC.in_out.graphics.type.context import (
    context_t,
    path_collection_h,
)
from cell_tracking_BC.in_out.graphics.type.event import event_h
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.s_viewer_2d import s_viewer_2d_t
from cell_tracking_BC.in_out.graphics.type.track import (
    VersionOfForkingForLayout,
)
from cell_tracking_BC.type.sequence import sequence_t
from cell_tracking_BC.type.track import forking_track_t, single_track_t


@dtcl.dataclass(repr=False, eq=False)
class t_viewer_2d_t:

    figure: figure_t
    axes_track: axes_t
    colormap: colormap_h
    viewer: s_viewer_2d_t

    scatter: path_collection_h = dtcl.field(init=False, default=None)
    annotation: annotation_h = dtcl.field(init=False, default=None)

    # Cell details
    labels: List[int] = dtcl.field(init=False, default_factory=list)
    time_points: List[int] = dtcl.field(init=False, default_factory=list)
    affinities: List[float] = dtcl.field(init=False, default_factory=list)
    colors: List[rgba_color_h] = dtcl.field(init=False, default_factory=list)

    dbe: context_t = None

    @classmethod
    def NewForSequence(
        cls,
        sequence: sequence_t,
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_segmentation: bool = True,
        with_cell_labels: bool = True,
        with_track_labels: bool = True,
        with_ticks: bool = True,
        with_colorbar: bool = True,
        figure_name: str = "tracking 2D",
        archiver: archiver_t = None,
    ) -> t_viewer_2d_t:
        """
        Annotation-on-hover code adapted from:
        https://stackoverflow.com/questions/7908636/how-to-add-hovering-annotations-in-matplotlib
        Answered by ImportanceOfBeingErnest on Nov 7 '17 at 20:23
        Edited by waterproof on Aug 12 '19 at 20:08
        """
        figure, two_axes = dbe.figure_2d_t.NewFigureAndAxes(n_cols=2)
        axes_track = cast(axes_t, two_axes[0])
        axes_viewer = cast(axes_t, two_axes[1])

        axes_track.SetTrackingAxesProperties(
            range(1, sum(_tck.n_leaves for _tck in sequence.tracks) + 1)
        )
        colormap = axes_track.AddStandardColormap(
            "Tracking Affinity", "plasma", position="left"
        )

        viewer = dbe.s_viewer_2d_t.NewForAllStreams(
            sequence,
            dbe,
            version=version,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            with_track_labels=with_track_labels,
            in_axes=axes_viewer,
            with_ticks=with_ticks,
            with_colorbar=with_colorbar,
        )
        # Cannot set conversion function here since it creates an import cycle
        # viewer.SetConversionToAnnotatedVolume(AsAnnotatedVolume)

        instance = cls(
            figure=figure,
            axes_track=axes_track,
            colormap=colormap,
            viewer=viewer,
            dbe=dbe,
        )

        all_cell_heights = []
        all_tracks = list(sequence.tracks)
        for which in (single_track_t, forking_track_t):
            if (invalids := sequence.tracks.invalids[which]) is not None:
                all_tracks.extend(_elm[0] for _elm in invalids)
        tick_details = []
        n_valid = sequence.tracks.__len__()
        for t_idx, track in enumerate(all_tracks):
            if isinstance(track, single_track_t):
                PlotTrackEdges = instance._PlotSingleTrackEdges
            else:
                PlotTrackEdges = instance._PlotForkingTrackEdges
            new_labels = PlotTrackEdges(track, all_cell_heights, t_idx < n_valid)
            tick_details.extend(new_labels)

        scatter = axes_track.PlotPoints(
            instance.time_points,
            all_cell_heights,
            marker="o",
            c=instance.colors,
            zorder=2,
        )
        positions, labels = zip(*tick_details)
        axes_track.SetAxisTicks("yaxis", positions, labels)
        annotation = axes_track.PlotAnnotation(
            "",
            xy=(0, 0),
            xytext=(20, 20),
            textcoords="offset points",
            bbox={"boxstyle": "round", "fc": "c"},
            arrowprops={"arrowstyle": "-"},
        )
        dbe.SetVisibility(annotation, False)

        instance.scatter = scatter
        instance.annotation = annotation

        figure.ActivateEvent("button_press_event", instance._OnButtonPress)

        figure.ActivateTightLayout(h_pad=0.05)
        figure.Archive(name=figure_name, archiver=archiver)

        return instance

    def _PlotSingleTrackEdges(
        self,
        track: single_track_t,
        all_cell_heights: List[int],
        is_valid: bool,
        /,
    ) -> Sequence[Tuple[float, int]]:
        """"""
        length, time_points, label, where = self._ElementsForTrackPieces(track)

        self.labels.insert(where, track.root.label)
        self.time_points.insert(where, track.root_time_point)
        self.affinities.insert(where, 0.0)
        self.colors.insert(where, self.colormap(0.0))

        heights = (length + 1) * (label,)
        all_cell_heights.extend(heights)

        if is_valid:
            color = "gray"
        else:
            color = "red"
        self.axes_track.PlotLines(time_points, heights, color=color, zorder=1)

        return ((label, label),)

    def _PlotForkingTrackEdges(
        self,
        track: forking_track_t,
        all_cell_heights: List[int],
        is_valid: bool,
        /,
    ) -> Sequence[Tuple[float, int]]:
        """"""
        with_int_labels, integer_to_cell = VersionOfForkingForLayout(track)
        try:
            int_layout = grph.nx_agraph.pygraphviz_layout(with_int_labels, prog="dot")
        except Exception as exc:
            wrng.warn(f"Track layout failed for {track} with error:\n{exc}")
            return ()
        positions = {
            integer_to_cell[_idx]: _pst
            for _idx, _pst in int_layout.items()
            if isinstance(_idx, int)
        }

        output = []

        all_time_points, all_heights = [], []
        min_height = max_height = positions[track.root][0]
        min_label = MAX_INTEGER
        root_height = None
        where = None
        for piece in track.Pieces():
            _, time_points, label, new_where = self._ElementsForTrackPieces(piece)
            heights = nmpy.fromiter(
                (positions[_cll][0] for _cll in piece), dtype=nmpy.float64
            )
            if piece[0] is track.root:
                root_height = heights[0]
            if where is None:
                where = new_where

            all_time_points.append(time_points)
            all_heights.append(heights)
            min_height = min(min_height, min(heights))
            max_height = max(max_height, max(heights))
            if label is not None:
                if label < min_label:
                    min_label = label
                output.append((heights[-1], label))

        height_scaling = (track.n_leaves - 1) / (max_height - min_height)
        AdjustedHeight = lambda _hgt: height_scaling * (_hgt - min_height) + min_label

        output = tuple((AdjustedHeight(_elm[0]), _elm[1]) for _elm in sorted(output))

        if is_valid:
            color = "gray"
        else:
            color = "red"
        for time_points, heights in zip(all_time_points, all_heights):
            heights = AdjustedHeight(heights)
            self.axes_track.PlotLines(time_points, heights, color=color, zorder=1)
            all_cell_heights.extend(heights[1:])
        root_height = AdjustedHeight(root_height)

        self.labels.insert(where, track.root.label)
        self.time_points.insert(where, track.root_time_point)
        self.affinities.insert(where, 0.0)
        self.colors.insert(where, self.colormap(0.0))

        all_cell_heights.insert(where, root_height)

        return output

    def _ElementsForTrackPieces(
        self,
        track: single_track_t,
        /,
    ) -> Tuple[int, Sequence[int], Optional[int], int]:
        """"""
        where = self.labels.__len__()

        label = track.label
        root_time_point = track.root_time_point
        length = track.length
        time_points = tuple(range(root_time_point, root_time_point + length + 1))
        affinities = track.affinities

        self.labels.extend(_cll.label for _cll in track[1:])
        self.time_points.extend(time_points[1:])
        self.affinities.extend(affinities)
        self.colors.extend(self.colormap(affinities))

        return length, time_points, label, where

    def _ShowAnnotation(
        self,
        event: event_h,
        /,
    ) -> None:
        """"""
        raise NotImplementedError

    def _OnButtonPress(
        self,
        event: event_h,
        /,
    ) -> None:
        """"""
        if self.dbe.IsTargetOfEvent(self.axes_track, event):
            self._ShowAnnotation(event)
        elif self.dbe.IsTargetOfEvent(self.viewer.axes, event):
            pass
        elif self.dbe.IsVisible(self.annotation):
            self.dbe.SetVisibility(self.annotation, False)
            self.figure.Update()

    def Show(
        self,
        /,
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        """"""
        self.figure.Show(interactively=interactively, in_main_thread=in_main_thread)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dcls
from abc import ABC as abc_t
from abc import abstractmethod
from typing import Iterator, Optional, Sequence, Tuple

import numpy as nmpy

import cell_tracking_BC.in_out.text.progress as prgs
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.tracks import (
    tracks_t,
    unstructured_tracks_t,
)


array_t = nmpy.ndarray


@dcls.dataclass(repr=False, eq=False)
class tracker_t(abc_t):
    """
    Latest tracking context: it contains all the information about the cells of frame t required to continue the
    tracking with the cells of frame t+1. When the tracks have been extended to the cells of frame t+1, this context
    will store the information about the cells of frame t+1 to get ready for continuing the tracking with the cells of
    frame t+2. This information is called the permanent storage of the context. By default, the permanent storage is
    defined as the attribute cells, a sequence of cell_t, and the temporary storage is defined as the attribute
    next_cells. Additional permanent and temporary attributes will generally be needed by contexts. Note that all these
    attributes go by pair since the method Advance must at least move all the temporary attributes to their permanent
    counterparts. The default implementation already move next_cells to cells. New contexts are in charge of moving the
    other attributes after calling super().Advance(). Similarly, the method DiscoverNextCells temporarily stores the
    discovered cells in next_cells, and new contexts are in charge of computing and storing the others attributes after
    calling super().DiscoverNextCells(next_cells) (See below for more details about permanent and temporary storages.)
    """

    cells: Sequence[cell_t] = dcls.field(init=False, default=None)
    next_cells: Sequence[cell_t] = dcls.field(init=False, default=None)

    def DiscoverNextCells(self, next_cells: Sequence[cell_t], /) -> None:
        """
        The tracking has been done up to frame t. This method is called when "discovering" the cells of frame t+1. It
        must "temporarily" store all the information about these cells required to extend the tracking.
        """
        self.next_cells = next_cells

    @abstractmethod
    def PredecessorOfCell(
        self, next_cell: cell_t, next_cell_idx: int, /
    ) -> Optional[Tuple[cell_t, float]]:
        """
        This method is called successively for all the cells of frame t+1 to extend the tracking track by track. It
        must look among the cells of frame t described in the permanent storage if one can be considered a predecessor
        of the given cell of frame t+1. If none, then it must return None, meaning that the current track ends.
        Otherwise, it must return the predecessor cell and an affinity measure for the cell association.
        """
        raise NotImplementedError(
            f"{tracker_t.PredecessorOfCell.__name__}: Must be implemented by trackers"
        )

    def Advance(self) -> None:
        """
        This method is called after PredecessorOfCell has been called for all the cells of frame t+1. It must move the
        temporary storage about the cells of frame t+1 created in DiscoverNextCells to the permanent storage that will
        be used in the next tracking iteration that will deal with frame t+2.
        """
        self.cells = self.next_cells


def CellsTracks(
    n_time_points: int,
    cells_per_time_point: Iterator[Sequence[cell_t]],
    tracker: tracker_t,
    /,
) -> tracks_t:
    """"""
    unstructured = unstructured_tracks_t()

    with prgs.ProgressDesign() as progress:
        progress_context = prgs.progress_context_t(
            progress,
            cells_per_time_point,
            total=n_time_points,
            description="Tracking Cells",
        )
        for t_idx, next_cells in enumerate(progress_context.elements):
            if next_cells.__len__() == 0:
                raise ValueError(f"No cells at time point {t_idx}")

            tracker.DiscoverNextCells(next_cells)

            if t_idx > 0:
                for next_c_idx, next_cell in enumerate(next_cells):
                    associated = tracker.PredecessorOfCell(next_cell, next_c_idx)
                    if associated is not None:
                        cell, affinity = associated
                        unstructured.AddTrackSegment(
                            cell, next_cell, t_idx - 1, affinity
                        )

            tracker.Advance()

    return tracks_t.NewFromUnstructuredTracks(unstructured)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from pathlib import Path as path_t
from typing import Any, Callable, Sequence, Tuple, Union

import matplotlib.pyplot as pypl
import numpy as nmpy
from matplotlib import cm as colormap_t
from matplotlib.backend_bases import Event as event_t
from matplotlib.colors import LinearSegmentedColormap as linear_colormap_t
from matplotlib.lines import Line2D as line_t
from matplotlib.text import Annotation as annotation_t
from matplotlib.widgets import Slider as slider_t

from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.type.axes import axes_t as base_axes_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t as base_figure_t
from cell_tracking_BC.in_out.text.logger import LOGGER, WhereMethod


array_t = nmpy.ndarray
pypl_axes_t = pypl.Axes
pypl_figure_t = pypl.Figure


_SHOWN_FIGURES = []


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_t(pypl_axes_t, base_axes_t):
    def SetTitle(self, title: str, /) -> None:
        """"""
        if self.axison:
            self.set_title(title)

    def SetAxisTicks(
        self, which: Union[str, Any], positions: Sequence[float], labels: Sequence[str], /
    ) -> None:
        """"""
        axis = getattr(self, which)
        axis.set_ticks(positions)
        axis.set_ticklabels(labels)

    def TurnTicksOff(
        self,
    ) -> None:
        """"""
        self.set_axis_off()

    def Freeze(self) -> None:
        """"""
        self.autoscale(enable=False)

    def ShowLegend(self) -> None:
        """"""
        self.legend()

    def AddStandardColormap(
        self, name: str, colormap: str, /, *, position: str = "right"
    ) -> colormap_t:
        """"""
        output = colormap_t.get_cmap(colormap)

        _AddColorbar(name, self.figure, self, output, position=position)

        return output

    def AddColormapFromMilestones(
        self,
        name: str,
        milestones: Sequence[Tuple[float, str]],
        /,
        *,
        position: str = "right",
    ) -> colormap_t:
        """"""
        output = linear_colormap_t.from_list(name, milestones)

        _AddColorbar(name, self.figure, self, output, position=position)

        return output

    def RemoveLinesAndAnnotations(self) -> None:
        """"""
        to_be_removed = filter(
            lambda _art: isinstance(_art, (annotation_t, line_t)),
            self.get_children(),
        )

        # tuple: to build a static list before iterative removal, just in case
        for annotation in tuple(to_be_removed):
            annotation.remove()

    def Figure(self) -> figure_t:
        """"""
        return self.figure


@dtcl.dataclass(init=False, repr=False, eq=False)
class figure_t(pypl_figure_t, base_figure_t):

    ActivateTightLayout = pypl_figure_t.tight_layout

    def Show(
        self,
        /,
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        """
        in_main_thread: disregarded
        """
        if self in _SHOWN_FIGURES:
            return

        self.show()
        _SHOWN_FIGURES.append(self)

        event_manager = self.canvas
        if interactively:
            if in_main_thread:
                event_manager.mpl_connect(
                    "close_event", lambda _: _StopEventLoop(event_manager, self)
                )
                event_manager.start_event_loop()
            else:
                LOGGER.warn(
                    f"{WhereMethod(self, self.Show)}: Interactivity is not possible outside the main thread"
                )
                # There seems to be no solution to have interactivity out of the main thread with Matplotlib
                event_manager.mpl_connect(
                    "close_event", lambda _: _SHOWN_FIGURES.remove(self)
                )
        else:
            event_manager.mpl_connect(
                "close_event", lambda _: _SHOWN_FIGURES.remove(self)
            )

    @staticmethod
    def ShowAll(
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        """"""
        if in_main_thread:
            pypl.show(block=interactively)
        else:
            for number in pypl.get_fignums():
                figure = pypl.figure(number)
                # Remember that in_main_thread is False
                figure.Show(interactively=interactively, in_main_thread=False)

    def Update(self, /, *, gently: bool = True) -> None:
        """"""
        if gently:
            self.canvas.draw_idle()
        else:
            self.canvas.draw()

    def Content(self, /) -> array_t:
        """
        Alternative:
            - fig.canvas.renderer._renderer?
            - Agg Buffer To Array: np.array(canvas.renderer.buffer_rgba())
        """
        height, width = self.canvas.get_width_height()
        output = nmpy.fromstring(self.canvas.tostring_rgb(), dtype=nmpy.uint8)
        output.shape = (width, height, 3)

        return output

    def Save(self, path: Union[str, path_t], /) -> None:
        """"""
        self.savefig(path, dpi=300)

    def Archive(
        self,
        /,
        *,
        name: str = "figure",
        archiver: archiver_t = None,
    ) -> None:
        """"""
        if archiver is not None:
            self.canvas.draw_idle()
            archiver.Store(self, name)

    def ActivateEvent(self, event: str, processor: Callable[[event_t], Any], /) -> None:
        """"""
        self.canvas.mpl_connect(event, processor)

    def Close(self, /) -> None:
        """"""
        pypl.close(fig=self)


def _StopEventLoop(event_manager: pypl.FigureCanvasBase, figure: figure_t, /) -> None:
    """"""
    event_manager.stop_event_loop()
    _SHOWN_FIGURES.remove(figure)


def _AddColorbar(
    name: str,
    figure: figure_t,
    axes: axes_t,
    colormap: colormap_t,
    /,
    *,
    position: str = "right",
) -> None:
    """"""
    mappable = colormap_t.ScalarMappable(cmap=colormap)
    figure.colorbar(mappable, ax=axes, location=position, label=name)


def NewSlider(figure: figure_t, n_steps: int, /) -> slider_t:
    """"""
    slider_axes = figure.add_axes([0.15, 0.04, 0.7, 0.03])

    return slider_t(
        slider_axes,
        "Time Point",
        0,
        n_steps - 1,
        valinit=0,
        valstep=1,
    )


def SliderValue(slider: slider_t, /) -> float:
    """"""
    return slider.val


def UpdateSlider(slider: slider_t, value: float, /) -> None:
    """"""
    slider.set_val(value)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# from __future__ import annotations
#
# import dataclasses as dtcl
# import warnings as wrng
# from sys import maxsize as MAX_INTEGER
# from typing import Dict, List, Optional, Sequence, Tuple, cast, Union
#
# import networkx as grph
# import numpy as nmpy
#
# from cell_tracking_BC.in_out.file.archiver import archiver_t
# from cell_tracking_BC.in_out.graphics.context import axes_2d_t as axes_t
# from cell_tracking_BC.in_out.graphics.context import (
#     context_t,
#     annotation_h,
#     event_h,
#     figure_t,
# )
# from cell_tracking_BC.in_out.graphics.generic.tree_layout import (
#     VersionOfForkingForLayout,
# )
# from cell_tracking_BC.type.track import forking_track_t, single_track_t
#
#
# array_t = nmpy.ndarray
#
#
# @dtcl.dataclass(repr=False, eq=False)
# class e_viewer_2d_t:
#
#     figure: figure_t
#     axes_track: axes_t
#
#     annotation: annotation_h = dtcl.field(init=False, default=None)
#
#     features: Sequence[str] = dtcl.field(init=False, default=None)
#     values: Sequence[Dict[int,array_t]] = dtcl.field(init=False, default=None)
#
#     dbe: context_t = None
#
#     @classmethod
#     def NewForFeature(
#         cls,
#         feature: Union[str, Sequence[str]],
#         values: Union[Dict[int,array_t], Sequence[Dict[int,array_t]]],
#         dbe: context_t,
#         /,
#         *,
#         figure_name: str = "feature",
#         archiver: archiver_t = None,
#     ) -> e_viewer_2d_t:
#         """
#         Annotation-on-hover code adapted from:
#         https://stackoverflow.com/questions/7908636/how-to-add-hovering-annotations-in-matplotlib
#         Answered by ImportanceOfBeingErnest on Nov 7 '17 at 20:23
#         Edited by waterproof on Aug 12 '19 at 20:08
#         """
#         figure, axes_track = dbe.figure_2d_t.NewFigureAndAxes()
#         axes_track = cast(axes_t, axes_track)
#
#         if isinstance(feature, str):
#             features = (feature,)
#             values = (values,)
#         else:
#             features = feature
#
#         max_label = max(values[0].keys())
#         axes_track.SetAxesPropertiesForTracking(
#             range(1, max_label+1, max(1,int(round(max_label / 10))))
#         )
#
#         instance = cls(
#             figure=figure,
#             axes_track=axes_track,
#             dbe=dbe,
#         )
#
#         rectangles = []
#         div_values = []
#         dea_values = []
#         for (label, division_response), death_response in zip(
#             cell_division_response[feature].items(), cell_death_response[feature].values()
#         ):
#             if division_response is None:
#                 continue
#
#             division_time_points = cell_division_frame_idc[feature][label]
#             death_frame_idx = cell_death_frame_idc[feature][label]
#
#             for t_idx, (div_value, dea_value) in enumerate(
#                 zip(division_response, death_response)
#             ):
#                 dividing = t_idx in division_time_points
#                 dying = t_idx == death_frame_idx
#                 rectangle = (
#                     rectangle_t((t_idx, label - 0.5), 1.0, 1.0, edgecolor="none"),
#                     (div_value, dea_value, dividing, dying),
#                 )
#                 rectangles.append(rectangle)
#             div_values.extend(division_response)
#             dea_values.extend(death_response)
#
#         if rectangles.__len__() > 0:
#             min_div_value = min(div_values)
#             max_div_value = max(div_values)
#             div_scaling = 1.0 / (max_div_value - min_div_value)
#             min_dea_value = min(dea_values)
#             max_dea_value = max(dea_values)
#             dea_scaling = 1.0 / (max_dea_value - min_dea_value)
#             events = []
#             for rectangle, (div_value, dea_value, dividing, dying) in rectangles:
#                 div = div_scaling * (div_value - min_div_value)
#                 dea = dea_scaling * (dea_value - min_dea_value)
#                 div_color = nmpy.array((1.0 - div, 1.0 - div, 1.0))
#                 dea_color = nmpy.array((1.0, 1.0 - dea, 1.0 - dea))
#                 color = nmpy.minimum(div_color, dea_color)
#                 rectangle.set_facecolor(color)
#                 if dividing or dying:
#                     where = tuple(_crd + 0.5 for _crd in rectangle.get_xy())
#                     if dividing:
#                         events.append(("3", 60, where))
#                     if dying:
#                         events.append(("x", 30, where))
#             figure, axes = pypl.subplots()
#             for rectangle, _ in rectangles:
#                 axes.add_artist(rectangle)
#             for what, size, where in events:
#                 axes.scatter(*where, c="k", marker=what, s=size)
#             max_label = max(cell_division_response[feature].keys())
#             axes.set_xlim((0, sequence.length))
#             axes.set_ylim((0,max_label + 1))
#             positions = range(1, max_label+1, max(1,int(round(max_label / 10))))
#             axes.yaxis.set_ticks(positions)
#             axes.yaxis.set_ticklabels(positions)
#             n_shades = 8
#             shades = nmpy.linspace(0.0, 1.0, num=n_shades)
#             div_legend = [
#                 rectangle_t(
#                     (0, 0),
#                     3.0,
#                     5.0,
#                     edgecolor="none",
#                     facecolor=(1.0 - _shd, 1.0 - _shd, 1.0),
#                 )
#                 for _shd in shades
#             ]
#             dea_legend = [
#                 rectangle_t(
#                     (0, 0),
#                     3.0,
#                     5.0,
#                     edgecolor="none",
#                     facecolor=(1.0, 1.0 - _shd, 1.0 - _shd),
#                 )
#                 for _shd in shades
#             ]
#             axes.legend(
#                 handles=(div_legend, dea_legend),
#                 labels=("Division Score", "Death Score"),
#                 loc="right",
#                 bbox_to_anchor=(1.3, 0.5),
#                 handler_map={list: tuple_handler_t(ndivide=None, pad=0)},
#             )
#
#         instance.features = features
#         instance.values = values
#
#         # figure.ActivateEvent("button_press_event", instance._OnButtonPress)
#
#         figure.ActivateTightLayout(h_pad=0.05)
#         figure.Archive(name=figure_name, archiver=archiver)
#
#         return instance
#
#     def _PlotSingleTrackEdges(
#         self,
#         track: single_track_t,
#         all_cell_heights: List[int],
#         is_valid: bool,
#         /,
#     ) -> Sequence[Tuple[float, int]]:
#         """"""
#         length, time_points, label, where = self._ElementsForTrackPieces(track)
#
#         self.labels.insert(where, track.root.label)
#         self.time_points.insert(where, track.root_time_point)
#         self.affinities.insert(where, 0.0)
#         self.colors.insert(where, self.colormap(0.0))
#
#         heights = (length + 1) * (label,)
#         all_cell_heights.extend(heights)
#
#         if is_valid:
#             color = "gray"
#         else:
#             color = "red"
#         self.axes_track.PlotLines(time_points, heights, color=color, zorder=1)
#
#         return ((label, label),)
#
#     def _PlotForkingTrackEdges(
#         self,
#         track: forking_track_t,
#         all_cell_heights: List[int],
#         is_valid: bool,
#         /,
#     ) -> Sequence[Tuple[float, int]]:
#         """"""
#         with_int_labels, integer_to_cell = VersionOfForkingForLayout(track)
#         try:
#             int_layout = grph.nx_agraph.pygraphviz_layout(with_int_labels, prog="dot")
#         except Exception as exc:
#             wrng.warn(f"Track layout failed for {track} with error:\n{exc}")
#             return ()
#         positions = {
#             integer_to_cell[_idx]: _pst
#             for _idx, _pst in int_layout.items()
#             if isinstance(_idx, int)
#         }
#
#         output = []
#
#         all_time_points, all_heights = [], []
#         min_height = max_height = positions[track.root][0]
#         min_label = MAX_INTEGER
#         root_height = None
#         where = None
#         for piece in track.Pieces():
#             _, time_points, label, new_where = self._ElementsForTrackPieces(piece)
#             heights = nmpy.fromiter(
#                 (positions[_cll][0] for _cll in piece), dtype=nmpy.float64
#             )
#             if piece[0] is track.root:
#                 root_height = heights[0]
#             if where is None:
#                 where = new_where
#
#             all_time_points.append(time_points)
#             all_heights.append(heights)
#             min_height = min(min_height, min(heights))
#             max_height = max(max_height, max(heights))
#             if label is not None:
#                 if label < min_label:
#                     min_label = label
#                 output.append((heights[-1], label))
#
#         height_scaling = (track.n_leaves - 1) / (max_height - min_height)
#         AdjustedHeight = lambda _hgt: height_scaling * (_hgt - min_height) + min_label
#
#         output = tuple((AdjustedHeight(_elm[0]), _elm[1]) for _elm in sorted(output))
#
#         if is_valid:
#             color = "gray"
#         else:
#             color = "red"
#         for time_points, heights in zip(all_time_points, all_heights):
#             heights = AdjustedHeight(heights)
#             self.axes_track.PlotLines(time_points, heights, color=color, zorder=1)
#             all_cell_heights.extend(heights[1:])
#         root_height = AdjustedHeight(root_height)
#
#         self.labels.insert(where, track.root.label)
#         self.time_points.insert(where, track.root_time_point)
#         self.affinities.insert(where, 0.0)
#         self.colors.insert(where, self.colormap(0.0))
#
#         all_cell_heights.insert(where, root_height)
#
#         return output
#
#     def _ElementsForTrackPieces(
#         self,
#         track: single_track_t,
#         /,
#     ) -> Tuple[int, Sequence[int], Optional[int], int]:
#         """"""
#         where = self.labels.__len__()
#
#         label = track.label
#         root_time_point = track.root_time_point
#         length = track.length
#         time_points = tuple(range(root_time_point, root_time_point + length + 1))
#         affinities = track.affinities
#
#         self.labels.extend(_cll.label for _cll in track[1:])
#         self.time_points.extend(time_points[1:])
#         self.affinities.extend(affinities)
#         self.colors.extend(self.colormap(affinities))
#
#         return length, time_points, label, where
#
#     def _ShowAnnotation(
#         self,
#         event: event_h,
#         /,
#     ) -> None:
#         """"""
#         raise NotImplementedError
#
#     def _OnButtonPress(
#         self,
#         event: event_h,
#         /,
#     ) -> None:
#         """"""
#         if self.dbe.IsTargetOfEvent(self.axes_track, event):
#             self._ShowAnnotation(event)
#         elif self.dbe.IsTargetOfEvent(self.viewer.axes, event):
#             pass
#         elif self.dbe.IsVisible(self.annotation):
#             self.dbe.SetVisibility(self.annotation, False)
#             self.figure.Update()
#
#     def Show(
#         self,
#         /,
#         *,
#         interactively: bool = True,
#         in_main_thread: bool = True,
#     ) -> None:
#         """"""
#         self.figure.Show(interactively=interactively, in_main_thread=in_main_thread)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from pathlib import Path as path_t
from typing import Any, Callable, Sequence, Tuple, Union

from numpy import ndarray as array_t

# Can import axes
from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.type.axes import axes_t
from cell_tracking_BC.in_out.graphics.type.event import event_h


@dtcl.dataclass(init=False, repr=False, eq=False)
class figure_t:
    @classmethod
    def NewFigureAndAxes(
        cls,
        /,
        *,
        n_rows: int = 1,
        n_cols: int = 1,
        title: str = None,
        offline_version: bool = False,
    ) -> Tuple[figure_t, Union[axes_t, Sequence[axes_t], Sequence[Sequence[axes_t]]]]:
        pass

    def ActivateTightLayout(self, *, pad=1.08, h_pad=None, w_pad=None, rect=None):
        pass

    def Show(
        self,
        /,
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        pass

    @staticmethod
    def ShowAll(
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        pass

    def Update(self, /, *, gently: bool = True) -> None:
        pass

    def Content(self, /) -> array_t:
        pass

    def Save(self, path: Union[str, path_t], /) -> None:
        pass

    def Archive(
        self,
        /,
        *,
        name: str = "figure",
        archiver: archiver_t = None,
    ) -> None:
        pass

    def ActivateEvent(self, event: str, processor: Callable[[event_h], Any], /) -> None:
        pass

    def Close(self) -> None:
        pass



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import matplotlib.pyplot as pypl
import numpy as nmpy
import skimage.data as data
import skimage.transform as trsf

import cell_tracking_BC.task.registration as rgst


angle = 18.0  # Must be between -20 and 20
print(f"true rotation={angle}")

image = data.retina()[..., 0]
threshold = nmpy.median(image)

image_as_int = (image > threshold).astype(nmpy.int8)

rotated = trsf.rotate(image_as_int, angle)
rotated_as_int = (rotated > 0.5 * nmpy.amax(rotated)).astype(nmpy.int8)

rotation = rgst.RotationInBinary(image_as_int, rotated_as_int)
print(f"{rotation=}")

rotated_back = rgst.RotatedLabeled(rotated_as_int, -rotation)

images = (image, image_as_int, rotated, rotated_as_int, rotated_back)
names = (
    "Original",
    "Thresholded",
    f"Rotated {angle}",
    "Rotated (int)",
    f"Rotated Back {-rotation}",
)
figure, all_axes = pypl.subplots(ncols=images.__len__())
for axes, image, name in zip(all_axes, images, names):
    axes.matshow(image, cmap="gray")
    axes.set_axis_off()
    axes.set_title(name)
figure.set_tight_layout(True)
pypl.show()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import sys as sstm
from typing import Iterable, Union

from rich.progress import BarColumn as progress_bar_t
from rich.progress import Progress as progress_t
from rich.progress import ProgressType as element_t
from rich.progress import TaskID as progress_id_t
from rich.progress import TimeElapsedColumn as time_elapsed_t
from rich.progress import TimeRemainingColumn as time_remaining_t


tracked_elements_t = Iterable[element_t]


@dtcl.dataclass(repr=False, eq=False)
class progress_context_t:
    progress: Union[progress_t, silent_progress_t]
    #
    last: dtcl.InitVar[Union[int, Iterable]]
    first: dtcl.InitVar[int] = 0
    total: dtcl.InitVar[int] = 0  # Used only if not isinstance(last, int)
    description: dtcl.InitVar[str] = "Progress"
    update_period: dtcl.InitVar[float] = 1.0
    #
    is_silent: bool = dtcl.field(init=False)
    uid: progress_id_t = dtcl.field(init=False)
    elements: tracked_elements_t = dtcl.field(init=False)

    def __post_init__(self, last, first, total, description, update_period) -> None:
        """"""
        if isinstance(last, int):
            iterable = range(first, last + 1)
            kwargs = {}
        else:
            iterable = last
            kwargs = {"total": total}

        self.is_silent = isinstance(self.progress, silent_progress_t)
        if self.is_silent:
            self.elements = iterable
        else:
            self.uid = self.progress.add_task(description)
            self.elements = self.progress.track(
                iterable, task_id=self.uid, update_period=update_period, **kwargs
            )
            sstm.stdout.flush()

    def UpdateDescription(self, description: str, /) -> None:
        """"""
        if self.is_silent:
            return

        self.progress.update(self.uid, description=description)


def ProgressDesign(
    *, should_be_silent: bool = False
) -> Union[progress_t, silent_progress_t]:
    """"""
    if should_be_silent:
        return silent_progress_t()

    return progress_t(
        "[progress.description]{task.description}",
        progress_bar_t(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        "[progress.elapsed]+",
        time_elapsed_t(),
        "/ [progress.remaining]-",
        time_remaining_t(),
    )


class silent_progress_t:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Optional

from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.context import DRAWING_CONTEXT
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t
from cell_tracking_BC.in_out.graphics.type.context import context_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.type.sequence import sequence_t


def ShowCellFeatureEvolution(
    sequence: sequence_t,
    feature: str,
    /,
    *,
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    figure_name: str = "segmentation",
    dbe: context_t = DRAWING_CONTEXT,
    archiver: archiver_t = None,
) -> Optional[figure_t]:
    """"""
    output = None

    figure, axes = dbe.figure_2d_t.NewFigureAndAxes(title=feature)
    axes: axes_2d_t

    evolutions = sequence.FeatureEvolutionsAlongAllTracks(feature)
    max_track_length = 0
    for label, (track, evolution) in evolutions.items():
        axes.PlotLines(
            range(track.root_time_point, track.leaf_time_point + 1),
            evolution,
            label=label,
        )

        if track.length > max_track_length:
            max_track_length = track.length
    axes.SetTimeAxisProperties(max_track_length)
    axes.ShowLegend()

    figure.Archive(name=figure_name, archiver=archiver)
    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Any, Dict, List, Sequence, Tuple

import numpy as nmpy
import skimage.morphology as mrph

import cell_tracking_BC.in_out.text.progress as prgs
from cell_tracking_BC.type.segmentation import (
    CellIssues_h,
    compartment_t,
    segmentation_t,
    version_id_h,
)


array_t = nmpy.ndarray


@dtcl.dataclass(repr=False, eq=False)
class segmentations_t(List[segmentation_t]):
    @classmethod
    def NewFromCompartmentSequences(
        cls,
        expected_length: int,
        /,
        *,
        cells_sgms: Sequence[array_t] = None,
        cytoplasms_sgms: Sequence[array_t] = None,
        nuclei_sgms: Sequence[array_t] = None,
    ) -> segmentations_t:
        """
        Valid options: see cell_tracking_BC.type.segmentation.segmentation_t.NewFromCompartments
        """
        instance = cls()

        fake_sequence = expected_length * [None]
        if cells_sgms is None:
            cells_sgms = fake_sequence
        if cytoplasms_sgms is None:
            cytoplasms_sgms = fake_sequence
        if nuclei_sgms is None:
            nuclei_sgms = fake_sequence
        all_sequences = (cells_sgms, cytoplasms_sgms, nuclei_sgms)
        lengths = tuple(map(len, all_sequences))
        if any(_lgt != expected_length for _lgt in lengths):
            raise ValueError(
                f"{lengths}: Non-none sequences do not all have expected length {expected_length}"
            )

        for cell, cytoplasm, nucleus in zip(*all_sequences):
            segmentation = segmentation_t.NewFromCompartments(
                cell=cell, cytoplasm=cytoplasm, nucleus=nucleus
            )
            instance.append(segmentation)

        return instance

    @classmethod
    def NewFromDicts(cls, dictionaries: Sequence[Dict[str, Any]]) -> segmentations_t:
        """"""
        instance = cls()

        for dictionary in dictionaries:
            segmentation = segmentation_t.NewFromDict(dictionary)
            instance.append(segmentation)

        return instance

    @property
    def available_versions(
        self,
    ) -> Tuple[Tuple[compartment_t, ...], Sequence[version_id_h]]:
        """
        See cell_tracking_BC.type.segmentation.available_versions
        """
        return self[0].available_versions

    def CompartmentsWithVersion(
        self, compartment: compartment_t, /, *, index: int = None, name: str = None
    ) -> Sequence[array_t]:
        """
        index: see cell_tracking_BC.type.segmentation.CompartmentWithVersion
        """
        output = []

        for segmentation in self:
            version = segmentation.CompartmentWithVersion(
                compartment, index=index, name=name
            )
            output.append(version)

        return output

    @property
    def cell_areas(self) -> Sequence[int]:
        """"""
        output = []

        for segmentation in self:
            labeled, n_cells = mrph.label(
                segmentation.cell, return_num=True, connectivity=1
            )
            areas = (
                nmpy.count_nonzero(labeled == _lbl) for _lbl in range(1, n_cells + 1)
            )
            output.extend(areas)

        return output

    def CellAreaHistogram(
        self,
        /,
        *,
        n_bins: int = None,
        should_return_centers: bool = False,
        should_round_centers: bool = True,
    ) -> Tuple[array_t, array_t]:
        """"""
        areas = self.cell_areas

        if n_bins is None:
            n_bins = int(round(nmpy.sqrt(areas.__len__())))
        counts, edges = nmpy.histogram(areas, bins=n_bins)
        if should_return_centers:
            centers = 0.5 * (edges[:-1] + edges[1:])
            if should_round_centers:
                centers = nmpy.around(centers).astype(nmpy.uint64)
            edges = centers

        return counts, edges

    def ClearBorderObjects(self) -> None:
        """"""
        for segmentation in self:
            segmentation.ClearBorderObjects()

    def FilterCellsOut(
        self,
        CellIssues: CellIssues_h,
        /,
        **kwargs,
    ) -> None:
        """
        Currently, only applicable to the cell segmentation when no other compartments are present

        Parameters
        ----------
        CellIssues: Arguments are: cell label (from 1), labeled segmentation, and (optional) keyword arguments; Returned
            value can be None, an str, or a sequence of str.
        kwargs: Passed to CellIsInvalid as keyword arguments with "frame_idx" (from 0) automatically added
        """
        if "frame_idx" in kwargs:
            raise ValueError(
                f'{kwargs}: Parameter name "frame_idx" '
                f"is reserved by {segmentations_t.FilterCellsOut.__name__}"
            )

        kwargs["frame_idx"] = None
        for f_idx, segmentation in enumerate(self):
            kwargs["frame_idx"] = f_idx
            segmentation.FilterCellsOut(CellIssues, **kwargs)

    def CorrectBasedOnTemporalCoherence(
        self,
        /,
        *,
        min_jaccard: float = 0.75,
        max_area_discrepancy: float = 0.25,
        min_cell_area: int = 0,
    ) -> None:
        """
        Actually, Pseudo-Jaccard
        """
        base_description = "Segmentation Correction(s) "
        n_corrections = 0
        with prgs.ProgressDesign() as progress:
            progress_context = prgs.progress_context_t(
                progress,
                self.__len__() - 1,
                first=1,
                description=base_description + "0",
            )
            for f_idx in progress_context.elements:
                n_corrections += self[f_idx].CorrectBasedOnTemporalCoherence(
                    self[f_idx - 1],
                    min_jaccard=min_jaccard,
                    max_area_discrepancy=max_area_discrepancy,
                    min_cell_area=min_cell_area,
                    time_point=f_idx,
                )
                progress_context.UpdateDescription(
                    base_description + str(n_corrections)
                )

        # Otherwise, the first frame does not have the same versions as the other ones
        self[0].AddFakeVersion(self[1].version_name)

    def AsDicts(self) -> Sequence[Dict[str, Any]]:
        """"""
        return [_sgm.AsDict() for _sgm in self]
